"""
    Copyright (C) 2013 Mayank Jha <mayank25080562@gmail.com>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
"""
    Copyright (C) 2013 Mayank Jha <mayank25080562@gmail.com>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 2.1 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui/fontCompare.ui'
#
# Created: Tue Nov 12 00:06:23 2013
#      by: PyQt4 UI code generator 4.10
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName(_fromUtf8("MainWindow"))
        MainWindow.resize(801, 600)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.verticalLayoutWidget = QtGui.QWidget(self.centralwidget)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(0, 0, 801, 561))
        self.verticalLayoutWidget.setObjectName(_fromUtf8("verticalLayoutWidget"))
        self.verticalLayout = QtGui.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setMargin(0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.groupBox = QtGui.QGroupBox(self.verticalLayoutWidget)
        self.groupBox.setObjectName(_fromUtf8("groupBox"))
        self.horizontalLayoutWidget_3 = QtGui.QWidget(self.groupBox)
        self.horizontalLayoutWidget_3.setGeometry(QtCore.QRect(9, 29, 781, 241))
        self.horizontalLayoutWidget_3.setObjectName(_fromUtf8("horizontalLayoutWidget_3"))
        self.horizontalLayout_3 = QtGui.QHBoxLayout(self.horizontalLayoutWidget_3)
        self.horizontalLayout_3.setMargin(0)
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        self.groupBox_7 = QtGui.QGroupBox(self.horizontalLayoutWidget_3)
        self.groupBox_7.setObjectName(_fromUtf8("groupBox_7"))
        self.layoutWidget_2 = QtGui.QWidget(self.groupBox_7)
        self.layoutWidget_2.setGeometry(QtCore.QRect(10, 50, 241, 170))
        self.layoutWidget_2.setObjectName(_fromUtf8("layoutWidget_2"))
        self.formLayout_3 = QtGui.QFormLayout(self.layoutWidget_2)
        self.formLayout_3.setFieldGrowthPolicy(QtGui.QFormLayout.ExpandingFieldsGrow)
        self.formLayout_3.setMargin(0)
        self.formLayout_3.setObjectName(_fromUtf8("formLayout_3"))
        self.label_18 = QtGui.QLabel(self.layoutWidget_2)
        self.label_18.setObjectName(_fromUtf8("label_18"))
        self.formLayout_3.setWidget(0, QtGui.QFormLayout.LabelRole, self.label_18)
        self.NumeralScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.NumeralScoreBar.setMaximum(10)
        self.NumeralScoreBar.setProperty("value", 0)
        self.NumeralScoreBar.setObjectName(_fromUtf8("NumeralScoreBar"))
        self.formLayout_3.setWidget(0, QtGui.QFormLayout.FieldRole, self.NumeralScoreBar)
        self.label_19 = QtGui.QLabel(self.layoutWidget_2)
        self.label_19.setObjectName(_fromUtf8("label_19"))
        self.formLayout_3.setWidget(1, QtGui.QFormLayout.LabelRole, self.label_19)
        self.ConsonantScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.ConsonantScoreBar.setMaximum(10)
        self.ConsonantScoreBar.setProperty("value", 0)
        self.ConsonantScoreBar.setObjectName(_fromUtf8("ConsonantScoreBar"))
        self.formLayout_3.setWidget(1, QtGui.QFormLayout.FieldRole, self.ConsonantScoreBar)
        self.label_20 = QtGui.QLabel(self.layoutWidget_2)
        self.label_20.setObjectName(_fromUtf8("label_20"))
        self.formLayout_3.setWidget(2, QtGui.QFormLayout.LabelRole, self.label_20)
        self.VowelScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.VowelScoreBar.setMaximum(10)
        self.VowelScoreBar.setProperty("value", 0)
        self.VowelScoreBar.setObjectName(_fromUtf8("VowelScoreBar"))
        self.formLayout_3.setWidget(2, QtGui.QFormLayout.FieldRole, self.VowelScoreBar)
        self.label_21 = QtGui.QLabel(self.layoutWidget_2)
        self.label_21.setObjectName(_fromUtf8("label_21"))
        self.formLayout_3.setWidget(3, QtGui.QFormLayout.LabelRole, self.label_21)
        self.MarkScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.MarkScoreBar.setMaximum(10)
        self.MarkScoreBar.setProperty("value", 0)
        self.MarkScoreBar.setObjectName(_fromUtf8("MarkScoreBar"))
        self.formLayout_3.setWidget(3, QtGui.QFormLayout.FieldRole, self.MarkScoreBar)
        self.label_22 = QtGui.QLabel(self.layoutWidget_2)
        self.label_22.setObjectName(_fromUtf8("label_22"))
        self.formLayout_3.setWidget(4, QtGui.QFormLayout.LabelRole, self.label_22)
        self.BasicConsistencyScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.BasicConsistencyScoreBar.setMaximum(10)
        self.BasicConsistencyScoreBar.setProperty("value", 0)
        self.BasicConsistencyScoreBar.setObjectName(_fromUtf8("BasicConsistencyScoreBar"))
        self.formLayout_3.setWidget(4, QtGui.QFormLayout.FieldRole, self.BasicConsistencyScoreBar)
        self.scriptRoundingLabel = QtGui.QLabel(self.layoutWidget_2)
        self.scriptRoundingLabel.setObjectName(_fromUtf8("scriptRoundingLabel"))
        self.formLayout_3.setWidget(5, QtGui.QFormLayout.LabelRole, self.scriptRoundingLabel)
        self.RoundingScoreBar = QtGui.QProgressBar(self.layoutWidget_2)
        self.RoundingScoreBar.setMaximum(10)
        self.RoundingScoreBar.setProperty("value", 0)
        self.RoundingScoreBar.setObjectName(_fromUtf8("RoundingScoreBar"))
        self.formLayout_3.setWidget(5, QtGui.QFormLayout.FieldRole, self.RoundingScoreBar)
        self.horizontalLayout_3.addWidget(self.groupBox_7)
        self.groupBox_6 = QtGui.QGroupBox(self.horizontalLayoutWidget_3)
        self.groupBox_6.setObjectName(_fromUtf8("groupBox_6"))
        self.layoutWidget = QtGui.QWidget(self.groupBox_6)
        self.layoutWidget.setGeometry(QtCore.QRect(10, 50, 241, 149))
        self.layoutWidget.setObjectName(_fromUtf8("layoutWidget"))
        self.formLayout_2 = QtGui.QFormLayout(self.layoutWidget)
        self.formLayout_2.setFieldGrowthPolicy(QtGui.QFormLayout.ExpandingFieldsGrow)
        self.formLayout_2.setMargin(0)
        self.formLayout_2.setObjectName(_fromUtf8("formLayout_2"))
        self.label_12 = QtGui.QLabel(self.layoutWidget)
        self.label_12.setObjectName(_fromUtf8("label_12"))
        self.formLayout_2.setWidget(0, QtGui.QFormLayout.LabelRole, self.label_12)
        self.highresScoreBar = QtGui.QProgressBar(self.layoutWidget)
        self.highresScoreBar.setMaximum(10)
        self.highresScoreBar.setProperty("value", 0)
        self.highresScoreBar.setObjectName(_fromUtf8("highresScoreBar"))
        self.formLayout_2.setWidget(0, QtGui.QFormLayout.FieldRole, self.highresScoreBar)
        self.label_13 = QtGui.QLabel(self.layoutWidget)
        self.label_13.setObjectName(_fromUtf8("label_13"))
        self.formLayout_2.setWidget(1, QtGui.QFormLayout.LabelRole, self.label_13)
        self.progressBar_10 = QtGui.QProgressBar(self.layoutWidget)
        self.progressBar_10.setMaximum(10)
        self.progressBar_10.setProperty("value", 0)
        self.progressBar_10.setObjectName(_fromUtf8("progressBar_10"))
        self.formLayout_2.setWidget(1, QtGui.QFormLayout.FieldRole, self.progressBar_10)
        self.label_14 = QtGui.QLabel(self.layoutWidget)
        self.label_14.setObjectName(_fromUtf8("label_14"))
        self.formLayout_2.setWidget(2, QtGui.QFormLayout.LabelRole, self.label_14)
        self.progressBar_11 = QtGui.QProgressBar(self.layoutWidget)
        self.progressBar_11.setMaximum(10)
        self.progressBar_11.setProperty("value", 0)
        self.progressBar_11.setObjectName(_fromUtf8("progressBar_11"))
        self.formLayout_2.setWidget(2, QtGui.QFormLayout.FieldRole, self.progressBar_11)
        self.label_16 = QtGui.QLabel(self.layoutWidget)
        self.label_16.setObjectName(_fromUtf8("label_16"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.LabelRole, self.label_16)
        self.progressBar_13 = QtGui.QProgressBar(self.layoutWidget)
        self.progressBar_13.setMaximum(10)
        self.progressBar_13.setProperty("value", 0)
        self.progressBar_13.setObjectName(_fromUtf8("progressBar_13"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.FieldRole, self.progressBar_13)
        self.horizontalLayout_3.addWidget(self.groupBox_6)
        self.groupBox_5 = QtGui.QGroupBox(self.horizontalLayoutWidget_3)
        self.groupBox_5.setObjectName(_fromUtf8("groupBox_5"))
        self.layoutWidget1 = QtGui.QWidget(self.groupBox_5)
        self.layoutWidget1.setGeometry(QtCore.QRect(10, 50, 241, 112))
        self.layoutWidget1.setObjectName(_fromUtf8("layoutWidget1"))
        self.formLayout = QtGui.QFormLayout(self.layoutWidget1)
        self.formLayout.setFieldGrowthPolicy(QtGui.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setMargin(0)
        self.formLayout.setObjectName(_fromUtf8("formLayout"))
        self.label_4 = QtGui.QLabel(self.layoutWidget1)
        self.label_4.setObjectName(_fromUtf8("label_4"))
        self.formLayout.setWidget(0, QtGui.QFormLayout.LabelRole, self.label_4)
        self.NormalScoreBar = QtGui.QProgressBar(self.layoutWidget1)
        self.NormalScoreBar.setMaximum(10)
        self.NormalScoreBar.setProperty("value", 0)
        self.NormalScoreBar.setObjectName(_fromUtf8("NormalScoreBar"))
        self.formLayout.setWidget(0, QtGui.QFormLayout.FieldRole, self.NormalScoreBar)
        self.label_5 = QtGui.QLabel(self.layoutWidget1)
        self.label_5.setObjectName(_fromUtf8("label_5"))
        self.formLayout.setWidget(1, QtGui.QFormLayout.LabelRole, self.label_5)
        self.BoldScoreBar = QtGui.QProgressBar(self.layoutWidget1)
        self.BoldScoreBar.setMaximum(10)
        self.BoldScoreBar.setProperty("value", 0)
        self.BoldScoreBar.setObjectName(_fromUtf8("BoldScoreBar"))
        self.formLayout.setWidget(1, QtGui.QFormLayout.FieldRole, self.BoldScoreBar)
        self.label_6 = QtGui.QLabel(self.layoutWidget1)
        self.label_6.setObjectName(_fromUtf8("label_6"))
        self.formLayout.setWidget(2, QtGui.QFormLayout.LabelRole, self.label_6)
        self.ItalicScoreBar = QtGui.QProgressBar(self.layoutWidget1)
        self.ItalicScoreBar.setMaximum(10)
        self.ItalicScoreBar.setProperty("value", 0)
        self.ItalicScoreBar.setObjectName(_fromUtf8("ItalicScoreBar"))
        self.formLayout.setWidget(2, QtGui.QFormLayout.FieldRole, self.ItalicScoreBar)
        self.label_7 = QtGui.QLabel(self.layoutWidget1)
        self.label_7.setObjectName(_fromUtf8("label_7"))
        self.formLayout.setWidget(3, QtGui.QFormLayout.LabelRole, self.label_7)
        self.BasicScoreBar = QtGui.QProgressBar(self.layoutWidget1)
        self.BasicScoreBar.setMaximum(10)
        self.BasicScoreBar.setProperty("value", 0)
        self.BasicScoreBar.setObjectName(_fromUtf8("BasicScoreBar"))
        self.formLayout.setWidget(3, QtGui.QFormLayout.FieldRole, self.BasicScoreBar)
        self.horizontalLayout_3.addWidget(self.groupBox_5)
        self.horizontalLayout.addWidget(self.groupBox)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.groupBox_3 = QtGui.QGroupBox(self.verticalLayoutWidget)
        self.groupBox_3.setObjectName(_fromUtf8("groupBox_3"))
        self.languageBox = QtGui.QComboBox(self.groupBox_3)
        self.languageBox.setGeometry(QtCore.QRect(250, 220, 141, 23))
        self.languageBox.setEditable(True)
        self.languageBox.setObjectName(_fromUtf8("languageBox"))
        self.languageBox.addItem(_fromUtf8(""))
        self.languageBox.addItem(_fromUtf8(""))
        self.languageBox.addItem(_fromUtf8(""))
        self.languageBox.addItem(_fromUtf8(""))
        self.BeginTestButton = QtGui.QPushButton(self.groupBox_3)
        self.BeginTestButton.setGeometry(QtCore.QRect(290, 30, 95, 71))
        self.BeginTestButton.setObjectName(_fromUtf8("BeginTestButton"))
        self.loadTestpushButton = QtGui.QPushButton(self.groupBox_3)
        self.loadTestpushButton.setGeometry(QtCore.QRect(10, 50, 181, 31))
        self.loadTestpushButton.setObjectName(_fromUtf8("loadTestpushButton"))
        self.label_2 = QtGui.QLabel(self.groupBox_3)
        self.label_2.setGeometry(QtCore.QRect(250, 200, 54, 13))
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.label_3 = QtGui.QLabel(self.groupBox_3)
        self.label_3.setGeometry(QtCore.QRect(10, 200, 51, 16))
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.comboBox = QtGui.QComboBox(self.groupBox_3)
        self.comboBox.setGeometry(QtCore.QRect(10, 220, 221, 22))
        self.comboBox.setEditable(True)
        self.comboBox.setObjectName(_fromUtf8("comboBox"))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.comboBox.addItem(_fromUtf8(""))
        self.label_17 = QtGui.QLabel(self.groupBox_3)
        self.label_17.setGeometry(QtCore.QRect(10, 180, 211, 16))
        self.label_17.setObjectName(_fromUtf8("label_17"))
        self.horizontalLayout_2.addWidget(self.groupBox_3)
        self.groupBox_4 = QtGui.QGroupBox(self.verticalLayoutWidget)
        self.groupBox_4.setObjectName(_fromUtf8("groupBox_4"))
        self.SaveMessageBoxButton = QtGui.QPushButton(self.groupBox_4)
        self.SaveMessageBoxButton.setGeometry(QtCore.QRect(20, 241, 91, 31))
        self.SaveMessageBoxButton.setObjectName(_fromUtf8("SaveMessageBoxButton"))
        self.ClearMessageBoxButton = QtGui.QPushButton(self.groupBox_4)
        self.ClearMessageBoxButton.setGeometry(QtCore.QRect(120, 241, 91, 31))
        self.ClearMessageBoxButton.setObjectName(_fromUtf8("ClearMessageBoxButton"))
        self.MessageBox = QtGui.QTextBrowser(self.groupBox_4)
        self.MessageBox.setGeometry(QtCore.QRect(15, 30, 371, 211))
        self.MessageBox.setObjectName(_fromUtf8("MessageBox"))
        self.horizontalLayout_2.addWidget(self.groupBox_4)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtGui.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 801, 25))
        self.menubar.setObjectName(_fromUtf8("menubar"))
        self.menu_File = QtGui.QMenu(self.menubar)
        self.menu_File.setObjectName(_fromUtf8("menu_File"))
        self.menuHelp = QtGui.QMenu(self.menubar)
        self.menuHelp.setObjectName(_fromUtf8("menuHelp"))
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setObjectName(_fromUtf8("statusbar"))
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoadTestFont = QtGui.QAction(MainWindow)
        self.actionLoadTestFont.setObjectName(_fromUtf8("actionLoadTestFont"))
        self.actionSaveTestResults = QtGui.QAction(MainWindow)
        self.actionSaveTestResults.setObjectName(_fromUtf8("actionSaveTestResults"))
        self.actionTest_Parameters = QtGui.QAction(MainWindow)
        self.actionTest_Parameters.setObjectName(_fromUtf8("actionTest_Parameters"))
        self.actionManual_Testing = QtGui.QAction(MainWindow)
        self.actionManual_Testing.setObjectName(_fromUtf8("actionManual_Testing"))
        self.actionAbout = QtGui.QAction(MainWindow)
        self.actionAbout.setObjectName(_fromUtf8("actionAbout"))
        self.actionabout = QtGui.QAction(MainWindow)
        self.actionabout.setObjectName(_fromUtf8("actionabout"))
        self.menu_File.addAction(self.actionLoadTestFont)
        self.menu_File.addAction(self.actionSaveTestResults)
        self.menuHelp.addAction(self.actionabout)
        self.menubar.addAction(self.menu_File.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(_translate("MainWindow", "Font Compare", None))
        self.groupBox.setTitle(_translate("MainWindow", "Categorised Scores", None))
        self.groupBox_7.setTitle(_translate("MainWindow", "Glyph Consistency", None))
        self.label_18.setText(_translate("MainWindow", "Numerals", None))
        self.NumeralScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_19.setText(_translate("MainWindow", "Consonants", None))
        self.ConsonantScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_20.setText(_translate("MainWindow", "Vowels", None))
        self.VowelScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_21.setText(_translate("MainWindow", "Marks", None))
        self.MarkScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_22.setText(_translate("MainWindow", "Basic Test", None))
        self.BasicConsistencyScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.scriptRoundingLabel.setText(_translate("MainWindow", "SRounding", None))
        self.RoundingScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.groupBox_6.setTitle(_translate("MainWindow", "Readability and Print Proof", None))
        self.label_12.setText(_translate("MainWindow", "High Resolution", None))
        self.highresScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_13.setText(_translate("MainWindow", "Low Resolution ", None))
        self.progressBar_10.setFormat(_translate("MainWindow", "%p", None))
        self.label_14.setText(_translate("MainWindow", "Low Light Read", None))
        self.progressBar_11.setFormat(_translate("MainWindow", "%p", None))
        self.label_16.setText(_translate("MainWindow", "Size(s) ", None))
        self.progressBar_13.setFormat(_translate("MainWindow", "%p", None))
        self.groupBox_5.setTitle(_translate("MainWindow", "Similarity to Standard font or Parameters", None))
        self.label_4.setText(_translate("MainWindow", "Normal", None))
        self.NormalScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_5.setText(_translate("MainWindow", "Bold ", None))
        self.BoldScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_6.setText(_translate("MainWindow", "Italic ", None))
        self.ItalicScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.label_7.setText(_translate("MainWindow", "Basic Properties", None))
        self.BasicScoreBar.setFormat(_translate("MainWindow", "%v", None))
        self.groupBox_3.setTitle(_translate("MainWindow", "Control Box", None))
        self.languageBox.setItemText(0, _translate("MainWindow", "All Range", None))
        self.languageBox.setItemText(1, _translate("MainWindow", "Devanagari", None))
        self.languageBox.setItemText(2, _translate("MainWindow", "Bengali ", None))
        self.languageBox.setItemText(3, _translate("MainWindow", "English", None))
        self.BeginTestButton.setText(_translate("MainWindow", "Begin Test!", None))
        self.loadTestpushButton.setText(_translate("MainWindow", "Load TestFont", None))
        self.label_2.setText(_translate("MainWindow", "Script:", None))
        self.label_3.setText(_translate("MainWindow", "Purpose:", None))
        self.comboBox.setItemText(0, _translate("MainWindow", "None", None))
        self.comboBox.setItemText(1, _translate("MainWindow", "Body Text : Books ", None))
        self.comboBox.setItemText(2, _translate("MainWindow", "Body Text: Magazines", None))
        self.comboBox.setItemText(3, _translate("MainWindow", "Body Text: Newspapers", None))
        self.comboBox.setItemText(4, _translate("MainWindow", "Body Text: Dictionaries", None))
        self.comboBox.setItemText(5, _translate("MainWindow", "Body Text: Directories", None))
        self.comboBox.setItemText(6, _translate("MainWindow", "Body Text: Greetings/Invitations", None))
        self.comboBox.setItemText(7, _translate("MainWindow", "Body Text: Poetry", None))
        self.comboBox.setItemText(8, _translate("MainWindow", "Body Text: Children Books", None))
        self.comboBox.setItemText(9, _translate("MainWindow", "Body Text: Advertisements", None))
        self.comboBox.setItemText(10, _translate("MainWindow", "Headline: Books", None))
        self.comboBox.setItemText(11, _translate("MainWindow", "Headline: Magazines", None))
        self.comboBox.setItemText(12, _translate("MainWindow", "Headline: Newspapers", None))
        self.comboBox.setItemText(13, _translate("MainWindow", "Headline: Advertisements", None))
        self.comboBox.setItemText(14, _translate("MainWindow", "Display: Posters/Signboards", None))
        self.comboBox.setItemText(15, _translate("MainWindow", "Display: Glow Signs", None))
        self.comboBox.setItemText(16, _translate("MainWindow", "Display: LED Display", None))
        self.comboBox.setItemText(17, _translate("MainWindow", "Display: TV/Mobiles Display", None))
        self.label_17.setText(_translate("MainWindow", "Additional Meta Information", None))
        self.groupBox_4.setTitle(_translate("MainWindow", "Test Results", None))
        self.SaveMessageBoxButton.setText(_translate("MainWindow", "Save To File", None))
        self.ClearMessageBoxButton.setText(_translate("MainWindow", "Clear All", None))
        self.menu_File.setTitle(_translate("MainWindow", "&File", None))
        self.menuHelp.setTitle(_translate("MainWindow", "Hel&p", None))
        self.actionLoadTestFont.setText(_translate("MainWindow", "Load TestFont", None))
        self.actionSaveTestResults.setText(_translate("MainWindow", "Save Test Results", None))
        self.actionTest_Parameters.setText(_translate("MainWindow", "Test Parameters", None))
        self.actionManual_Testing.setText(_translate("MainWindow", "Manual Testing", None))
        self.actionAbout.setText(_translate("MainWindow", "About", None))
        self.actionabout.setText(_translate("MainWindow", "About", None))

