((nil . ((eval . (c-add-style "fontforge"
                              '("stroustrup"
                                (indent-tabs-mode . t)
                                (tab-width . 8)
                                (c-offsets-alist
                                 (case-label . *)))))))
 (c-mode . ((c-file-style . "fontforge"))))
